from jovian._version import __version__
from jovian.utils.commit import commit
from jovian.utils.records import (log_hyperparams, log_metrics, log_dataset,
                                  log_git, reset, get_records)
from jovian.utils.slack import notify
from jovian.utils.submit import submit
from jovian.utils.configure import reset_config, configure
from jovian.utils.initialize import _initialize_jovian
from jovian.utils.misc import version

_initialize_jovian()



from jovian.utils.misc import urljoin
from jovian.utils.credentials import read_api_url


def _u(path):
    """Make a URL from the path"""
    return urljoin(read_api_url(), path)


def _v(version):
    """Create version query parameter string"""
    if version is not None:
        return "?gist_version=" + str(version)
    return ""




class ApiError(Exception):
    """Error class for web API related Exceptions"""
    pass


class ConfigError(Exception):
    """Error class for config related Exceptions"""
    pass


class CondaError(Exception):
    """Error class for Anaconda-related exceptions"""
    pass



from random import random

import requests
from pkg_resources import parse_version

from jovian._version import __version__
from jovian.utils.jupyter import in_notebook
from jovian.utils.logger import log


def _get_latest_version():
    """Returns Jovian's latest pypi version.

    When Internet connection is not available it assumes 
    the current version as the latest version.
    """

    pypi_json = 'https://pypi.org/pypi/jovian/json'
    try:
        latest_version = requests.get(pypi_json, timeout=3).json()['info']['version']
    except:
        latest_version = __version__

    return latest_version


def _print_update_message(current_version, latest_version):
    log('Update Available: {0} --> {1}'.format(current_version, latest_version))

    if in_notebook():
        log('Run `!pip install jovian --upgrade` to upgrade')
    else:
        log('Run `pip install jovian --upgrade` to upgrade\n')


def check_update(probability=0.8):
    """Check if there is a update available and logs
    current and latest version with the command to update
    the library.

    Args:
        probability (float, optional): Approximate probability with which the user
                                    is notified. Defaults to .2(~20%).
                                    (1 == always)
    """
    if random() < probability:
        latest_version = parse_version(_get_latest_version())
        current_version = parse_version(__version__)
        if latest_version > current_version:
            _print_update_message(current_version, latest_version)



from jovian.utils.api import _h
from jovian.utils.commit import commit, _parse_filename
from jovian.utils.credentials import read_webapp_url
from jovian.utils.error import ApiError
from jovian.utils.logger import log
from jovian.utils.misc import urljoin
from jovian.utils.request import post, pretty
from jovian.utils.shared import _u

POST_API = '/learn/course/section/{}/make_submission'
ASSIGNMENT_PAGE_URL = '/learn/{}/assignment/{}'


def submit(assignment=None, notebook_url=None, **kwargs):
    """ Performs jovian.commit and makes a assignment submission with the uploaded notebook.
    """
    if not assignment:
        log("Please provide assignment name", error=True)
        return

    filename = _parse_filename(kwargs.get('filename'))
    if filename == '__notebook_source__.ipynb':
        log("""jovian.submit does not support kaggle notebooks directly.
         Please make a commit first, copy the notebook URL and pass it to jovian.submit.
         eg. jovian.submit(assignment="zero-to-pandas-a1", 
                           notebook_url="https://jovian.ai/PrajwalPrashanth/assignment")""", error=True)
        return

    post_url = POST_API.format(assignment)
    nb_url = notebook_url if notebook_url else commit(**kwargs)

    if nb_url:
        data = {
            'assignment_url': nb_url
        }
        auth_headers = _h()

        log('Submitting assignment..')
        res = post(url=_u(post_url),
                   json=data,
                   headers=auth_headers)

        if res.status_code == 200:
            data = res.json()['data']
            course_slug = data.get('course_slug')
            assignment_slug = data.get('section_slug')

            assignment_page_url = ASSIGNMENT_PAGE_URL.format(course_slug, assignment_slug)
            log('Verify your submission at {}'.format(urljoin(read_webapp_url(), assignment_page_url)))
        else:
            log('Jovian submit failed. {}'.format(pretty(res)), error=True)



from __future__ import print_function
import click


def log(msg, pre=True, error=False, warn=False, color=None):
    """Print a message to stdout"""
    if error:
        click.secho(('[jovian] ' if pre else '') + 'Error: ' + msg, err=True, fg='red')
    elif warn:
        click.secho(('[jovian] ' if pre else '') + msg, fg='yellow')
    else:
        click.echo(('[jovian] ' if pre else '') + click.style(msg, fg=color))



# URLs for public versions of Jovian
DEFAULT_WEBAPP_URL = "https://jovian.ai/"
DEFAULT_API_URL = "https://api.jovian.ai/"

RC_FILENAME = ".jovianrc"
FILENAME_MSG = 'Failed to detect notebook filename. Please provide the correct notebook filename ' + \
    'as the "filename" argument to "jovian.commit".'
ISSUES_MSG = """NOTE: Jovian is currently in beta, so if you face any issues, 
               please report them here: https://github.com/JovianML/jovian-py/issues"""

LINUX = 'linux'
WINDOWS = 'windows'
MACOS = 'macos'
PLATFORMS = [LINUX, WINDOWS, MACOS]

STORAGE_ALLOWED_EXTENSIONS = tuple('ipynb yml yaml py txt csv'.split())

DEFAULT_ORG_ID = "public"

CONDA_NOT_FOUND = 'Anaconda binary not found. Please make sure the "conda" command is in your ' \
                  'system PATH or the environment variable $CONDA_EXE points to the anaconda binary'

DEFAULT_EXTENSION_WHITELIST = [".ipynb", ".yml", ".yaml", ".py", ".txt", ".csv", ".tsv"]



from jovian._version import __version__
from jovian.utils.credentials import get_api_key, get_guest_key, read_api_url, read_org_id
from jovian.utils.error import ApiError
from jovian.utils.logger import log
from jovian.utils.misc import timestamp_ms
from jovian.utils.request import get, post, pretty
from jovian.utils.shared import _u, _v


def _h():
    """Create authorization header with API key"""
    return {"Authorization": "Bearer " + get_api_key(),
            "x-jovian-source": "library",
            "x-jovian-library-version": __version__,
            "x-jovian-guest": get_guest_key(),
            "x-jovian-org": read_org_id()}


def get_current_user():
    res = get(url=_u('/user/profile'), headers=_h())
    if res.status_code == 200:
        return res.json()['data']
    raise Exception('Failed to fetch current user profile. ' + pretty(res))


def get_gist(slug, version=None, check_exists=True):
    """Get the metadata for a gist"""
    if '/' in slug:
        parts = slug.split('/')
        username, title = parts[0], parts[1]
        url = _u('user/' + username + '/gist/' + title + _v(version))
    else:
        url = _u('gist/' + slug + _v(version))
    res = get(url=url, headers=_h())
    if res.status_code == 200:
        return res.json()['data']
    elif check_exists and res.status_code == 404:
        return False
    raise Exception('Failed to retrieve metadata for notebook "' +
                    slug + '": ' + pretty(res))


def get_gist_access(slug):
    """Get the access permission of a gist"""
    res = get(url=_u('/gist/' + slug + '/check-access'), headers=_h())
    if res.status_code == 200:
        return res.json()['data']
    raise Exception('Failed to retrieve access permission for notebook "' +
                    slug + '" (retry with new_project=True to create a new notebook): ' + pretty(res))


def create_gist_simple(filename=None, gist_slug=None, privacy='auto', title=None, version_title=None):
    """Upload the current notebook to create/update a gist"""
    auth_headers = _h()

    with open(filename, 'rb') as f:
        nb_file = (filename, f)
        log('Uploading notebook..')
        if gist_slug:
            return upload_file(gist_slug=gist_slug, file=nb_file, version_title=version_title)
        else:
            data = {'visibility': privacy}

            # For compatibility with old version of API endpoint
            if privacy == 'auto':
                data['public'] = True
            elif privacy == 'secret' or privacy == 'private':
                data['public'] = False

            if title:
                data['title'] = title
            if version_title:
                data['version_title'] = version_title
            res = post(url=_u('/gist/create'),
                       data=data,
                       files={'files': nb_file},
                       headers=auth_headers)
            if res.status_code == 200:
                return res.json()['data']
            raise ApiError('File upload failed: ' + pretty(res))


def upload_file(gist_slug, file, folder=None, version=None, artifact=False, version_title=None):
    """Upload an additional file to a gist"""
    data = {'artifact': 'true'} if artifact else {}
    if folder:
        data['folder'] = folder
    if version_title:
        data['version_title'] = version_title

    res = post(url=_u('/gist/' + gist_slug + '/upload' + _v(version)),
               files={'files': file}, data=data, headers=_h())
    if res.status_code == 200:
        return res.json()['data']
    raise ApiError('File upload failed: ' + pretty(res))


def post_blocks(blocks, version=None):
    url = _u('/data/record' + _v(version))
    res = post(url, json=blocks, headers=_h())
    if res.status_code == 200:
        return res.json()['data']
    else:
        raise ApiError('Data logging failed: ' + pretty(res))


def post_block(data, data_type, version=None):
    """Upload metrics, hyperparameters and other information to server"""
    blocks = [{"localTimestamp": timestamp_ms(),
               "data": data,
               "recordType": data_type}]
    return post_blocks(blocks, version)


def post_records(gist_slug, tracking_slugs, version=None):
    """Associated tracked records with a commit"""
    url = _u('/data/' + gist_slug + '/commit' + _v(version))
    res = post(url, json=tracking_slugs, headers=_h())
    if res.status_code == 200:
        return res.json()['data']
    else:
        raise ApiError('Data logging failed: ' + pretty(res))


def post_slack_message(data, safe=False):
    """Push data to Slack, if slack is integrated with jovian account"""
    url = _u('/slack/notify')
    res = post(url, json=data, headers=_h())
    if res.status_code == 200:
        return res.json()
    elif safe:
        return {'data': {'messageSent': False}}
    else:
        raise ApiError('Slack trigger failed: ' + pretty(res))



import os
import sys
import time
import platform

from uuid import UUID
from jovian.utils.constants import LINUX, WINDOWS, MACOS
from jovian._version import __version__


def is_uuid(text):
    """Check if the given string is a UUID"""
    try:
        _ = UUID(text, version=4)
        return True
    except ValueError:
        return False


def is_py2():
    """Check if current Python version is 2.x"""
    return sys.version_info[0] < 3


def get_platform():
    """Identify which OS the library is running on

    Returns 'linux', 'windows' or 'macos'
    """
    system = platform.system()
    if system == 'Linux':
        return LINUX
    elif system == 'Windows':
        return WINDOWS
    elif system == 'Darwin':
        return MACOS


def timestamp_ms():
    """Return the current timestamp (in milliseconds)"""
    return int(time.time() * 1000)


def get_flavor():
    """Get the flavor of the library (jovian or jovian-pro)"""
    try:
        from jovian._flavor import __flavor__
        return __flavor__
    except ImportError:
        return "jovian"


def is_flavor_pro():
    """Get the flavor of the library (jovian or jovian-pro)"""
    return get_flavor() == 'jovian-pro' or get_flavor() == 'jovianpro'


def get_file_extension(filename):
    """Get the extension of a file"""
    return os.path.splitext(filename)[1]


def urljoin(*args):
    """Join multiple url parts to construct one url"""
    if len(args) == 0:
        raise TypeError("urljoin requires at least one argument")

    trailing_slash = '/' if args[-1].endswith('/') else ''

    return '/'.join(map(lambda x: str(x).strip('/'), args)) + trailing_slash


def version():
    return __version__



"""Utilities to install packages for a cloned gist"""
from __future__ import print_function
import subprocess
from time import sleep
from sys import stderr
from jovian.utils.environment import get_conda_bin, CONDA_NOT_FOUND, print_conda_message
from jovian.utils.constants import ISSUES_MSG
from jovian.utils.logger import log
from jovian.utils.envfile import (check_error, check_pip_failed, extract_env_name,
                                  extract_env_packages, extract_pip_packages,
                                  identify_env_file, request_env_name, sanitize_envfile)


def run_command(command, env_fname, packages, run=1):
    # Run the command
    if run > 3:
        return

    log('Executing:\n' + command + "\n")
    install_task = subprocess.Popen(command, shell=True, stderr=subprocess.PIPE)
    # Extract the error (if any)
    _, error_string = install_task.communicate()
    error_string = error_string.decode('utf8', errors='ignore')
    if error_string:
        print(error_string, file=stderr)
        # Check for errors
        error, pkgs = check_error(error_string, packages=packages)
        pip_failed = check_pip_failed(error_string)

        if error:
            log('Installation failed!', error=True)
            log('Ignoring ' + error + ' dependencies and trying again...\n')
            sleep(1)
            sanitize_envfile(env_fname=env_fname, pkgs=pkgs)
            return run_command(command=command, env_fname=env_fname, packages=packages, run=run+1)

        elif pip_failed:
            # TODO: Extract env details and run pip sub-command.
            # pip_packages = extract_pip_packages(env_fname=env_fname)
            return False
    else:
        # Print beta warning and github link
        log(ISSUES_MSG)
        return True


def install(env_fname=None, env_name=None):
    """Install packages for a cloned gist"""
    # Check for conda and get the binary path
    conda_bin = get_conda_bin()

    # Identify the right environment file, and exit if absent
    env_fname = identify_env_file(env_fname=env_fname)
    if env_fname is None:
        log('Failed to detect a conda environment YML file. Skipping..', error=True)
        return
    else:
        log('Detected conda environment file: ' + env_fname + "\n")

    # Get the environment name from user input
    env_name = request_env_name(env_name=env_name, env_fname=env_fname)
    if env_name is None:
        log('Environment name not provided/detected. Skipping..')
        return

    # Construct the command
    command = conda_bin + ' env update --file "' + \
        env_fname + '" --name "' + env_name + '"'

    packages = extract_env_packages(env_fname=env_fname)
    if len(packages) > 0:
        success = run_command(command=command, env_fname=env_fname, packages=packages, run=1)
        if not success:
            log('Some pip packages failed to install.')
            print_conda_message(env_name=env_name)


def activate(env_fname=None):
    """Read the conda environment file and activate the environment"""
    log(ISSUES_MSG)

    # Check for conda and get the binary path
    try:
        conda_bin = get_conda_bin()
    except:
        log(CONDA_NOT_FOUND, error=True)
        return False

    # Identify the right environment file, and exit if absent
    env_fname = identify_env_file(env_fname=env_fname)
    if env_fname is None:
        log('Failed to detect a conda environment YML file. Skipping..', error=True)
        return False
    else:
        log('Detected conda environment file: ' + env_fname + "\n")

    # Get the environment name from user input
    env_name = extract_env_name(env_fname=env_fname)
    if env_name is None:
        log('Environment name not provided/detected. Skipping..')
        return False

    # Activate the environment
    command = conda_bin + " activate " + env_name
    log('Copy and execute the following command (try "source activate" if "conda activate doesn\'t work" )')
    print("    " + command + " \n")



import json
from time import sleep


from jovian.utils.logger import log
from jovian.utils.api import get_current_user
from jovian.utils.credentials import read_cred, WEBAPP_URL_KEY
from jovian.utils.constants import DEFAULT_WEBAPP_URL


def perform_kaggle_commit(message,
                          files,
                          outputs,
                          environment,
                          privacy,
                          project,
                          new_project):
    """ Retreive all cells and writes it to a file called project-name.ipynb, then returns the filename"""
    from IPython import get_ipython

    # Get user profile
    user = get_current_user()['username']

    # Construct and print URL
    url = read_cred(WEBAPP_URL_KEY, DEFAULT_WEBAPP_URL) + user + "/" + project
    log("Uploading notebook to " + url)

    # Construct filename
    filename = "{}.ipynb".format(project.split('/')[1] if '/' in project else project)

    # Handle str, None accordingly
    message = "'{}'".format(message) if isinstance(message, str) else None
    environment = "'{}'".format(environment) if isinstance(environment, str) else None

    # Construct jovian.commit w/ parameters
    jovian_commit = """jovian.commit(message={}, files={}, outputs={}, environment={}, privacy='{}', filename='{}', project='{}', new_project={})""".format(
        message, files, outputs, environment, privacy, filename, project, new_project)

    # Construct javascript code
    js_code = '''
    require(["base/js/namespace"],function(Jupyter) {
        var nbJson = JSON.stringify(Jupyter.notebook.toJSON());

        console.log("[jovian] Extracted notebook JSON:");
        console.log(nbJson);

        function jvnLog (data) {
          console.log("Result from jovian.commit:");
          if (data.content.text) {
              var result = JSON.parse(data.content.text.trim());
              var msg = result['msg'];
              var err = result['err'];
              if (msg) {
                  element.text("Committed successfully: " + msg)
              } else {
                  alert("Notebook commit failed. Error: " + (err || "Unknown"))
              }
          }
          
        };
        
        var pythonCode = `
from contextlib import redirect_stdout, redirect_stderr
from io import StringIO
import json
 
with open("'''+filename+'''", 'w') as f:
    f.write(r"""${nbJson}""")

jvn_update = StringIO()
jvn_update_err = StringIO()
with redirect_stdout(jvn_update), redirect_stderr(jvn_update_err):
    from jovian import commit

jvn_f_out = StringIO()
jvn_f_err = StringIO()
with redirect_stdout(jvn_f_out), redirect_stderr(jvn_f_err):
    jvn_msg = '''+jovian_commit+'''

print(json.dumps({'msg': jvn_msg, 'err': jvn_f_err.getvalue(), 'update': jvn_update.getvalue()}))
        `;

        console.log("Invoking jovian.commit")
        // console.log(pythonCode)

        Jupyter.notebook.kernel.execute(pythonCode, { iopub: { output: jvnLog }});
    });'''

    get_ipython().run_cell_magic('javascript', '', js_code)



import click

from jovian.utils.api import get_api_key
from jovian.utils.credentials import creds_exist, ensure_org, get_guest_key, purge_creds
from jovian.utils.logger import log


def reset_config(confirm=True):
    """Remove the existing configuration by purging credentials"""
    if creds_exist():
        if confirm:
            msg = 'Do you want to remove the existing configuration?'
            confirmed = click.confirm(msg)
        else:
            confirmed = True

        if confirmed:
            log('Removing existing configuration. Run "jovian configure" to set up Jovian')
            purge_creds()
        else:
            log('Skipping..')
            return
    else:
        log('Jovian is not configured yet. Run "jovian configure" to set it up.')


def configure():
    """Configure Jovian for first time usage"""
    # Check if already exists
    if creds_exist():
        log('It looks like Jovian is already configured ( check ~/.jovian/credentials.json ).')
        msg = 'Do you want to overwrite the existing configuration?'
        confirm = click.confirm(msg)

        if confirm:
            log('Removing existing configuration..')
        else:
            log('Skipping..')
            return

    # Remove existing credentials
    purge_creds()

    # Capture and save organization ID
    ensure_org(check_pro=False)

    # Ask for API Key
    get_guest_key()
    get_api_key()

    log('Configuration complete!')



"""Anaconda and Pip related utilities"""
import os
import logging
from jovian.utils.api import upload_file
from jovian.utils.misc import get_platform
from jovian.utils.constants import PLATFORMS, CONDA_NOT_FOUND
from jovian.utils.logger import log
from jovian.utils.error import CondaError


def get_conda_bin():
    """Get the path to the Anaconda binary"""
    conda_bin = 'conda'
    # Try executing the `conda` command
    if os.popen(conda_bin).read().strip() == '':
        # If it fails, look for $CONDA_EXE
        conda_exe = os.environ.get("CONDA_EXE")
        # Check if it returns a valid path
        if conda_exe and conda_exe != '$CONDA_EXE':
            # Update binary and execute again
            conda_bin = conda_exe
            if os.popen(conda_bin).read().strip() == '':
                raise CondaError(CONDA_NOT_FOUND)
    logging.info('Anaconda binary: ' + conda_bin)
    return conda_bin


def get_conda_env_name():
    """Get the name of the active conda environment"""
    env_name = os.environ.get("CONDA_DEFAULT_ENV", "base")
    if not env_name:
        env_name = "base"
    logging.info('Anaconda environment: ' + env_name)
    return env_name


def read_conda_env(name=None):
    """Read the anaconda environment into a string"""
    if name is None:
        name = get_conda_env_name()
    command = get_conda_bin() + ' env export -n ' + name + " --no-builds"
    env_str = os.popen(command).read()
    if env_str == '':
        error = 'Failed to read Anaconda environment using command: "' + command + '"'
        raise CondaError(error)
    return env_str


def upload_conda_env(gist_slug, version=None):
    """Read and save the current Anaconda environment to server"""
    # Export environment to YML string
    env_str = read_conda_env(get_conda_env_name())

    # Upload environment.yml
    upload_file(gist_slug=gist_slug, file=('environment.yml', env_str), version=version)

    # Check and include existing os-specific files
    platform = get_platform()
    for p in PLATFORMS:
        pfname = 'environment-' + p + '.yml'
        if p == platform:
            # Use the new environment for current platform
            upload_file(gist_slug=gist_slug, file=(pfname, env_str), version=version)
        elif os.path.exists(pfname):
            # Reuse old environments for other platforms
            with open(pfname, 'rb') as f:
                file = (pfname, f)
                upload_file(gist_slug=gist_slug, file=file, version=version)


def print_conda_message(env_name):
    if env_name:
        message = ("""
#
# To activate this environment, use
#
#     $ conda activate %s
#
# To deactivate an active environment, use
#
#     $ conda deactivate
        """) % env_name
        log(message)


def read_pip_env():
    """Read the pip dependencies into a string"""
    command = "pip --disable-pip-version-check freeze"
    deps_str = os.popen(command).read()
    if deps_str == '':
        error = 'Failed to read pip environment using command: "' + command + '"'
        raise Exception(error)
    return deps_str


def upload_pip_env(gist_slug, version=None):
    """Read and upload the current virtual environment to server"""
    return upload_file(gist_slug=gist_slug, file=('requirements.txt', read_pip_env()), version=version)



import os
import re
import setuptools

VERSIONFILE = "./jovian/_version.py"
FLAVORFILE = "./jovian/_flavor.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
pkg_name = os.getenv("PKG_NAME", "jovian")

if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

with open(FLAVORFILE, "w") as f:
    f.write("__flavor__ = \"{}\"".format(pkg_name))


with open("README.md", "rb") as fh:
    long_description = fh.read().decode('utf-8', errors='ignore')

setuptools.setup(
    name=pkg_name,
    version=verstr,
    author="Jovian",
    author_email="hello@jovian.ai",
    entry_points={
        'console_scripts': ['jovian=jovian.__main__:main'],
    },
    description="Jovian Python SDK",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://jovian.ai/?utm_source=pypi",
    packages=setuptools.find_packages(exclude=["*.tests", "*.tests.*"]),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    package_data={"jovian": ['jovian_nb_ext/*']},
    data_files=[("share/jupyter/nbextensions/jovian_nb_ext", ["jovian/jovian_nb_ext/main.js"]),
                ("etc/jupyter/nbconfig/notebook.d", ["jovian/jovian_nb_ext/jovian_nb_ext.json"])],
    include_package_data=True,
    install_requires=['requests', 'uuid', 'pyyaml', 'click']
)



import glob
import os
import json
from time import sleep

import click
from jovian.utils import api, git
from jovian.utils.colab import in_colab, get_colab_file_id, perform_colab_commit
from jovian.utils.constants import DEFAULT_EXTENSION_WHITELIST, FILENAME_MSG
from jovian.utils.credentials import read_creds, read_webapp_url
from jovian.utils.environment import CondaError, upload_conda_env, upload_pip_env
from jovian.utils.jupyter import get_notebook_name, in_notebook, save_notebook
from jovian.utils.kaggle import perform_kaggle_commit
from jovian.utils.logger import log
from jovian.utils.misc import get_file_extension, is_uuid, urljoin
from jovian.utils.rcfile import get_cached_slug, get_notebook_slug, reset_notebook_slug, set_notebook_slug
from jovian.utils.records import get_records, log_git, reset
from jovian.utils.script import get_script_filename, in_script


def commit(message=None,
           files=[],
           outputs=[],
           environment='auto',
           privacy='auto',
           filename=None,
           project=None,
           new_project=None,
           git_commit=False,
           git_message='auto',
           **kwargs):
    """Uploads the current file (Jupyter notebook or python script) to |Jovian|

    Saves the checkpoint of the notebook, captures the required dependencies from
    the python environment and uploads the notebook, env file, additional files like scripts, csv etc.
    to |Jovian|. Capturing the python environment ensures that the notebook can be reproduced.

    Args:
        message(string, optional): A short message to be used as the title for this version.

        files(array, optional): Any additional scripts(.py files), CSVs etc. that are required to
            run the notebook. These will be available in the files tab of the project page on Jovian

        outputs(array, optional): Any outputs files or artifacts generated from the modeling processing.
            This can include model weights/checkpoints, generated CSVs, output images etc.

        environment(string, optional): The type of Python environment to be captured.  Allowed options are
            'conda' , 'pip', 'auto' (for automatic detection) and None (to skip environment capture).

        privacy(bool, optional): Privacy level of the project (if a new one is being created).

            * 'auto' - use account level settings. Defaults to 'public'
            * 'public' - visible on profile and publicly accessible/searchable
            * 'secret' - not on profile only accessible via the direct link
            * 'private' - only for the accessible to owner and collaborators

            This argument has no effect on existing project. Change the privacy settings of a existing notebook
            on the webapp.

        filename(string, optional): The filename of the current Jupyter notebook or Python script. This is
            detected automatically in most cases, but in certain environments like Jupyter Lab or password protected notebooks, the detection
            may fail and the filename needs to be provided using this argument.


        project(string, optional): Name of the |Jovian| project to which the current notebook/file should
            be committed. Format: 'username/title' e.g. 'aakashns/jovian-example' or 'jovian-example'
            (username of current user inferred automatically). If the project does not exist, a new one is
            created. If it exists, the current notebook is added as a new version to the existing project, if
            you are a owner/collaborator. If left empty, project name is picked up from the `.jovianrc` file in the
            current directory, or a new project is created using the filename as the project name.

        new_project(bool, optional): Whether to create a new project or update the existing one. Allowed option
            are False (use the existing project, if a .jovianrc file exists, if available), True (create a new project)

        git_commit(bool, optional): If True, also performs a Git commit and records the commit hash. This is
            applicable only when the notebook is inside a Git repository.

        git_message(string, optional): Commit message for git. If not provided, it uses the `message` argument

    .. attention::
        Pass notebook's name to `filename` argument, in certain environments like Jupyter Lab and password protected
        notebooks sometimes it may fail to detect notebook automatically.
    .. |Jovian| raw:: html

        <a href="https://jovian.ai/?utm_source=docs" target="_blank"> Jovian </a>
    """

    # Deprecated argument (secret)
    if privacy == 'auto' and 'secret' in kwargs:
        privacy = 'secret' if kwargs['secret'] else 'auto'
        log('"secret" is deprecated. Use "privacy" instead (allowed options: ' +
            '"public", "private", "secret", "auto")', error=True)

    # Deprecated argument (nb_filename)
    if filename is None and 'nb_filename' in kwargs:
        filename = kwargs['nb_filename']
        log('"nb_filename" is deprecated. Use "filename" instead', error=True)

    # Deprecated argument (env_type)
    if 'env_type' in kwargs:
        environment = kwargs['env_type']
        log('"env_type" is deprecated. Use "environment" instead', error=True)

    # Deprecated argument (capture_env)
    if 'capture_env' in kwargs and not kwargs['capture_env']:
        environment = None
        log('"catpure_env" is deprecated. Use "environment=None" instead', error=True)

    # Deprecated argument (notebook_id)
    if 'notebook_id' in kwargs:
        project = kwargs['notebook_id']
        log('"notebook_id" is deprecated. Use "project" instead.', error=True)

    # Deprecated argument (create_new)
    if 'create_new' in kwargs:
        new_project = kwargs['create_new']
        log('"create_new" is deprecated. Use "new_project" instead.', error=True)

    # Deprecated argument (artifacts)
    if 'artifacts' in kwargs:
        outputs = kwargs['artifacts']
        log('"artifacts" is deprecated. Use "outputs" instead', error=True)

    is_cli = kwargs.get('is_cli', False)

    # To commit colab notebooks
    if in_colab():
        log("Detected Colab notebook...")
        if project is None:
            log("Please provide the project argument e.g. jovian.commit(project='my-project')", error=True)
            return
        if get_colab_file_id() is None:
            log("Colab File Id is not provided. Make sure to execute the cell where the Id is set.", error=True)
            return

        res = perform_colab_commit(project, privacy)
        slug, username, version, title = res['slug'], res['owner']['username'], res['version'], res['title']

        if environment == "conda":
            log("Conda environment is not supported on Colab. Capturing pip environment instead.")
            environment = "pip"
        elif environment == "auto":
            environment = "pip"

        _capture_environment(environment, slug, version)
        _attach_files(files, slug, version)
        _attach_files(outputs, slug, version, output=True)
        _attach_records(slug, version)

        log('Committed successfully! ' + urljoin(read_webapp_url(), username, title))
        return urljoin(read_webapp_url(), username, title)

    # Skip if unsupported environment
    if not in_script() and not in_notebook() and not is_cli:
        log('Failed to detect Jupyter notebook or Python script. Skipping..', error=True)
        return

    # Attempt to save Jupyter notebook
    if in_notebook():
        save_notebook()
        log('Attempting to save notebook..')
        sleep(1)

    # Extract notebook/script filename
    filename = _parse_filename(filename)
    if filename is None:
        log(FILENAME_MSG, error=True)
        return

    # Commit from Kaggle (After many bug reports of empty notebook)
    if filename == '__notebook_source__.ipynb':
        log("Detected Kaggle notebook...")
        if not project:
            log("Please provide the project argument e.g. jovian.commit(project='my-project')", error=True)
            return

        perform_kaggle_commit(message,
                              files,
                              outputs,
                              environment,
                              privacy,
                              project,
                              new_project)
        return

    # Ensure that the file exists
    if not os.path.exists(filename):
        log('The detected/provided file "' + filename +
            '" does not exist. Please provide the correct notebook filename ' +
            'as the "filename" argument to "jovian.commit".', error=True)
        return

    # Retrieve Gist ID & title
    project_title, project_id = _parse_project(project, filename, new_project)

    # Create or update gist (with title and )
    res = api.create_gist_simple(filename, project_id, privacy, project_title, message)
    slug, owner, version, title = res['slug'], res['owner'], res['version'], res['title']
    username = owner['username']

    # Cache slug for further commits
    set_notebook_slug(filename, slug)

    # Attach environment, files and outputs
    _capture_environment(environment, slug, version)
    _attach_files(files, slug, version, exclude_files=filename)
    _attach_files(outputs, slug, version, output=True)

    if not git_message or git_message == 'auto':
        git_message = message or 'jovian commit ' + username + '/' + title + ' v' + str(version)
    _perform_git_commit(filename, git_commit, git_message)
    _attach_records(slug, version)

    log('Committed successfully! ' + urljoin(read_webapp_url(), username, title))

    return urljoin(read_webapp_url(), username, title)


def commit_path(path, **kwargs):
    files = _list_ipynb_files(path)
    num_files = len(files)

    if num_files == 0:
        log("No notebook found in path: " + path)
        return

    if num_files >= 50:
        log("Can't upload more than 50 files at once")
        return

    if num_files == 1:
        log('Uploading 1 notebook: {}'.format(files[0]))
    else:
        log('Uploading {} notebooks:'.format(num_files))
        log('\n'.join(files), pre=False)

    if click.confirm('\n[jovian] Do you want to continue?', default=True):
        for filename in files:
            reset_notebook_slug()
            commit(filename=filename, **kwargs)
            log("", pre=False)
            sleep(1)


def _parse_filename(filename):
    """Perform the required checks and get the final filename"""
    # Get the filename of the notebook (if not provided)
    if filename is None:
        if in_script():
            filename = get_script_filename()
        elif in_notebook():
            filename = get_notebook_name()

    # Add the right extension to the filename
    elif get_file_extension(filename) not in ['.py', '.ipynb']:
        filename += '.py' if in_script() else '.ipynb'
    return filename


def _parse_project(project, filename, new_project):
    """Perform the required checks and get the final project name"""
    current_slug = get_cached_slug()

    # Check for existing project in-memory or in .jovianrc
    if not new_project and project is None:
        # From in-memory variable
        if current_slug is not None:
            project = current_slug
        # From .jovianrc file
        else:
            project = get_notebook_slug(filename)

    # Skip if project is not provided & can't be read
    if project is None:
        return None, None

    # Get project metadata for UUID & username/title
    if is_uuid(project):
        project_title = None
        metadata = api.get_gist(project)
    elif '/' in project:
        project_title = project.split('/')[1]
        username = api.get_current_user()['username']
        metadata = api.get_gist(project)
    # Attach username to the title
    else:
        project_title = project
        username = api.get_current_user()['username']
        metadata = api.get_gist(username + '/' + project)

    # Skip if metadata could not be found
    if not metadata:
        log('Creating a new project "' + username + '/' + project_title + '"')
        return project_title, None

    # Extract information from metadata
    username = metadata['owner']['username']
    project_title = metadata['title']
    project_id = metadata['slug']

    # Check if the current user can commit to this project
    permissions = api.get_gist_access(project_id)
    if not permissions['write']:
        return project_title, None

    # Log whether this is an update or creation
    if project_id is None:
        log('Creating a new notebook on ' + read_webapp_url())
    else:
        log('Updating notebook "' + username + "/" + project_title + '" on ' + read_webapp_url())

    return project_title, project_id


def _attach_file(path, gist_slug, version, output=False):
    """Helper function to attach a single file to a commit"""
    try:
        with open(path, 'rb') as f:
            file_obj = os.path.basename(path), f
            folder = os.path.dirname(path)
            api.upload_file(gist_slug, file_obj, folder, version, output)
    except Exception as e:
        log(str(e) + " (" + path + ")", error=True)


def _attach_files(paths, gist_slug, version, output=False, exclude_files=None):
    """Helper functions to attach files & folders to a commit"""
    config = read_creds().get("DEFAULT_CONFIG", {})

    whitelist = config.get("EXTENSION_WHITELIST")
    upload_wd = config.get("UPLOAD_WORKING_DIRECTORY", False)

    if not isinstance(whitelist, list):
        whitelist = DEFAULT_EXTENSION_WHITELIST

    if not paths:
        if output or not upload_wd:
            return

        paths = [
            f
            for f in glob.glob('**/*', recursive=True)
            if get_file_extension(f) in whitelist
        ]

    if exclude_files:
        if not isinstance(exclude_files, list):
            exclude_files = [exclude_files]

        for filename in exclude_files:
            try:
                paths.remove(filename)
            except ValueError:
                pass

    log('Uploading additional ' + ('outputs' if output else 'files') + '...')

    # Convert single path to list
    if type(paths) == str:
        paths = [paths]

    for path in paths:
        if os.path.isdir(path):
            files = [
                f
                for f in glob.glob(os.path.join(path, '**/*'), recursive=True)
                if get_file_extension(f) in whitelist
            ]
            for file in files:
                _attach_file(file, gist_slug, version, output)
        elif os.path.exists(path):
            _attach_file(path, gist_slug, version, output)
        else:
            log('Ignoring "' + path + '" (not found)', error=True)


def _capture_environment(environment, gist_slug, version):
    """Capture the python environment and attach it to the commit"""
    if environment is not None:
        # Check credentials if environment config exists
        creds = read_creds()
        if 'DEFAULT_CONFIG' in creds and 'environment' in creds['DEFAULT_CONFIG']:
            environment_config = creds['DEFAULT_CONFIG']['environment']
            if not environment_config:
                # Disable environment capture
                return
            if environment == 'auto' and (environment_config == 'conda' or environment_config == 'pip'):
                environment = environment_config

        log('Capturing environment..')
        captured = False

        if environment == 'auto' or environment == 'conda':
            # Capture conda environment
            try:
                upload_conda_env(gist_slug, version)
                captured = True
            except CondaError as e:
                log(str(e), error=True)

        if not captured and (environment == 'pip' or environment == 'auto'):
            # Capture pip environment
            try:
                upload_pip_env(gist_slug, version)
            except Exception as e:
                log(str(e), error=True)


def _perform_git_commit(filename, git_commit, git_message):
    if git_commit and git.is_git():
        reset('git')  # resets git commit info

        git.commit(git_message)
        log('Git repository identified. Performing git commit...')

        git_info = {
            'repository': git.get_remote(),
            'commit': git.get_current_commit(),
            'filename': filename,
            'path': git.get_relative_path(),
            'branch': git.get_branch()
        }
        log_git(git_info, verbose=False)


def _attach_records(gist_slug, version):
    """Attached records to the current commit"""
    tracking_slugs = get_records(slug_only=True)
    if len(tracking_slugs) > 0:
        log('Attaching records (metrics, hyperparameters, dataset etc.)')
        api.post_records(gist_slug, tracking_slugs, version)


def _list_ipynb_files(path):
    """Return list of ipynb files in a path"""
    if os.path.isfile(path) and get_file_extension(path) == ".ipynb":
        files = [os.path.normpath(path)]
    elif os.path.isdir(path):
        files = [os.path.normpath(f) for f in sorted(glob.glob(os.path.join(path, '*.ipynb')))]
    else:
        files = []
    return files
